module_list = {
    "raster": {
        "cost surface": {
            "r.cost": {
                "keywords": "raster,cost surface,cumulative costs,cost "
                "allocation",
                "description": "Creates a raster map showing the cumulative "
                "cost of  moving between different geographic locations on an "
                "input raster map whose cell category values represent"
                " cost.",
                "module": "r.cost",
            },
            "r.walk": {
                "keywords": "raster,cost surface,cumulative costs,cost "
                "allocation",
                "description": "Creates a raster map showing the anisotropic "
                "cumulative cost of moving between different geographic "
                "locations on an input raster map whose cell category values "
                "represent cost.",
                "module": "r.walk",
            },
        },
        "projection": {
            "r.proj": {
                "keywords": "raster,projection,transformation,import",
                "description": "Re-projects a raster map from given location "
                "to the current location.",
                "module": "r.proj",
            }
        },
        "transect": {
            "r.transect": {
                "keywords": "raster,transect",
                "description": "Outputs raster map layer values lying along "
                "user defined transect line(s).",
                "module": "r.transect",
            }
        },
        "series": {
            "r.series.interp": {
                "keywords": "raster,series,interpolation",
                "description": "Interpolates raster maps located (temporal or "
                "spatial) in between input raster maps at specific sampling "
                "positions.",
                "module": "r.series.interp",
            },
            "r.series.accumulate": {
                "keywords": "raster,series,accumulation",
                "description": "Makes each output cell value a "
                "accumulationfunction of the values assigned to the "
                "corresponding cells in the input raster map layers.",
                "module": "r.series.accumulate",
            },
        },
        "terrain": {
            "r.slope.aspect": {
                "keywords": "raster,terrain,aspect,slope,curvature",
                "description": "Generates raster maps of slope, aspect, "
                "curvatures and partial derivatives from an elevation raster "
                "map. Aspect is calculated counterclockwise from east.",
                "module": "r.slope.aspect",
            }
        },
        "surface": {
            "r.surf.area": {
                "keywords": "raster,surface,statistics,area estimation",
                "description": "Prints estimation of surface area for raster "
                "map.",
                "module": "r.surf.area",
            },
            "r.surf.fractal": {
                "keywords": "raster,surface,fractal",
                "description": "Creates a fractal surface of a given fractal "
                "dimension.",
                "module": "r.surf.fractal",
            },
            "r.surf.contour": {
                "keywords": "raster,surface,interpolation",
                "description": "Generates surface raster map from rasterized "
                "contours.",
                "module": "r.surf.contour",
            },
            "r.resamp.bspline": {
                "keywords": "raster,surface,resample,interpolation",
                "description": "Performs bilinear or bicubic spline "
                "interpolation with Tykhonov regularization.",
                "module": "r.resamp.bspline",
            },
            "r.surf.idw": {
                "keywords": "raster,surface,interpolation,IDW",
                "description": "Provides surface interpolation from raster "
                "point data by Inverse Distance Squared Weighting.",
                "module": "r.surf.idw",
            },
            "r.surf.gauss": {
                "keywords": "raster,surface,random",
                "description": "Generates a raster map using gaussian random "
                "number generator. Mean and standard deviation of gaussian "
                "deviates can be expressed by the user.",
                "module": "r.surf.gauss",
            },
            "r.random.surface": {
                "keywords": "raster,surface,random",
                "description": "Generates random surface(s) with spatial "
                "dependence.",
                "module": "r.random.surface",
            },
            "r.contour": {
                "keywords": "raster,surface,contours,vector",
                "description": "Produces a vector map of specified contours "
                "from a raster map.",
                "module": "r.contour",
            },
            "r.surf.random": {
                "keywords": "raster,surface,random",
                "description": "Produces a raster surface map of uniform "
                "random deviates with defined range.",
                "module": "r.surf.random",
            },
        },
        "sampling": {
            "r.random.cells": {
                "keywords": "raster,sampling,random,autocorrelation",
                "description": "Generates random cell values with spatial "
                "dependence.",
                "module": "r.random.cells",
            },
            "r.random": {
                "keywords": "raster,sampling,vector,random",
                "description": "Creates a raster map layer and vector point "
                "map containing randomly located points.",
                "module": "r.random",
            },
        },
        "RGB": {
            "r.rgb": {
                "keywords": "raster,RGB,separate,split",
                "description": "Splits a raster map into red, green and blue "
                "maps.",
                "module": "r.rgb",
            }
        },
        "recode categories": {
            "r.recode": {
                "keywords": "raster,recode categories,reclassification",
                "description": "Recodes categorical raster maps.",
                "module": "r.recode",
            }
        },
        "landscape structure analysis": {
            "r.li.dominance": {
                "keywords": "raster,landscape structure analysis,diversity "
                "index",
                "description": "Calculates dominance's diversity index on a "
                "raster map",
                "module": "r.li.dominance",
            },
            "r.li.mpa": {
                "keywords": "raster,landscape structure analysis,patch index",
                "description": "Calculates mean pixel attribute index on a "
                "raster map",
                "module": "r.li.mpa",
            },
            "r.li.padcv": {
                "keywords": "raster,landscape structure analysis,patch index",
                "description": "Calculates coefficient of variation of patch "
                "area on a raster map",
                "module": "r.li.padcv",
            },
            "r.li.patchnum": {
                "keywords": "raster,landscape structure analysis,patch index",
                "description": "Calculates patch number index on a raster map,"
                " using a 4 neighbour algorithm.",
                "module": "r.li.patchnum",
            },
            "r.li.padsd": {
                "keywords": "raster,landscape structure analysis,patch index",
                "description": "Calculates standard deviation of patch area a "
                "raster map",
                "module": "r.li.padsd",
            },
            "r.li.cwed": {
                "keywords": "raster,landscape structure analysis,patch index",
                "description": "Calculates contrast weighted edge density "
                "index on a raster map",
                "module": "r.li.cwed",
            },
            "r.li.edgedensity": {
                "keywords": "raster,landscape structure analysis,patch index",
                "description": "Calculates edge density index on a raster map,"
                " using a 4 neighbour algorithm",
                "module": "r.li.edgedensity",
            },
            "r.li.simpson": {
                "keywords": "raster,landscape structure analysis,diversity "
                "index",
                "description": "Calculates Simpson's diversity index on a "
                "raster map",
                "module": "r.li.simpson",
            },
            "r.li.renyi": {
                "keywords": "raster,landscape structure analysis,diversity "
                "index",
                "description": "Calculates Renyi's diversity index on a raster"
                " map",
                "module": "r.li.renyi",
            },
            "r.li.mps": {
                "keywords": "raster,landscape structure analysis,patch index",
                "description": "Calculates mean patch size index on a raster "
                "map, using a 4 neighbour algorithm",
                "module": "r.li.mps",
            },
            "r.li.shannon": {
                "keywords": "raster,landscape structure analysis,diversity "
                "index",
                "description": "Calculates Shannon's diversity index on a "
                "raster map",
                "module": "r.li.shannon",
            },
            "r.li.patchdensity": {
                "keywords": "raster,landscape structure analysis,patch index",
                "description": "Calculates patch density index on a raster "
                "map, using a 4 neighbour algorithm",
                "module": "r.li.patchdensity",
            },
            "r.li.richness": {
                "keywords": "raster,landscape structure analysis,diversity "
                "index",
                "description": "Calculates richness index on a raster map",
                "module": "r.li.richness",
            },
            "r.li.pielou": {
                "keywords": "raster,landscape structure analysis,diversity "
                "index",
                "description": "Calculates Pielou's diversity index on a "
                "raster map",
                "module": "r.li.pielou",
            },
            "r.li.shape": {
                "keywords": "raster,landscape structure analysis,patch index",
                "description": "Calculates shape index on a raster map",
                "module": "r.li.shape",
            },
            "r.li.padrange": {
                "keywords": "raster,landscape structure analysis,patch index",
                "description": "Calculates range of patch area size on a "
                "raster map",
                "module": "r.li.padrange",
            },
        },
        "tiling": {
            "r.tile": {
                "keywords": "raster,tiling",
                "description": "Splits a raster map into tiles.",
                "module": "r.tile",
            },
            "r.tileset": {
                "keywords": "raster,tiling",
                "description": "Produces tilings of the source projection for "
                "use in the destination region and projection.",
                "module": "r.tileset",
            },
        },
        "groundwater flow": {
            "r.gwflow": {
                "keywords": "raster,groundwater flow,hydrology",
                "description": "Numerical calculation program for transient, "
                "confined and unconfined groundwater flow in two dimensions.",
                "module": "r.gwflow",
            }
        },
        "reclassification": {
            "r.reclass": {
                "keywords": "raster,reclassification",
                "description": "Reclassify raster map based on category "
                "values. Creates a new raster map whose category values are "
                "based upon a reclassification of the categories in "
                "an existing raster map.",
                "module": "r.reclass",
            }
        },
        "category": {
            "r.category": {
                "keywords": "raster,category",
                "description": "Manages category values and labels associated "
                "with user-specified raster map layers.",
                "module": "r.category",
            }
        },
        "conversion": {
            "r.to.vect": {
                "keywords": "raster,conversion,geometry,vectorization",
                "description": "Converts a raster map into a vector map.",
                "module": "r.to.vect",
            },
            "r.to.rast3": {
                "keywords": "raster,conversion,voxel",
                "description": "Converts 2D raster map slices to one 3D raster"
                " volume map.",
                "module": "r.to.rast3",
            },
            "r.to.rast3elev": {
                "keywords": "raster,conversion,raster3d,voxel",
                "description": "Creates a 3D volume map based on 2D elevation "
                "and value raster maps.",
                "module": "r.to.rast3elev",
            },
        },
        "statistics": {
            "r.stats.zonal": {
                "keywords": "raster,statistics,zonal statistics",
                "description": "Calculates category or object oriented "
                "statistics (accumulator-based statistics).",
                "module": "r.stats.zonal",
            },
            "r.coin": {
                "keywords": "raster,statistics",
                "description": "Tabulates the mutual occurrence (coincidence) "
                "of categories for two raster map layers.",
                "module": "r.coin",
            },
            "r.support.stats": {
                "keywords": "raster,statistics",
                "description": "Update raster map statistics",
                "module": "r.support.stats",
            },
            "r.clump": {
                "keywords": "raster,statistics,reclass,clumps",
                "description": "Recategorizes data in a raster map by grouping"
                " cells that form physically discrete areas into unique "
                "categories.",
                "module": "r.clump",
            },
            "r.cross": {
                "keywords": "raster,statistics",
                "description": "Creates a cross product of the category values"
                " from multiple raster map layers.",
                "module": "r.cross",
            },
            "r.reclass.area": {
                "keywords": "raster,statistics,aggregation",
                "description": "Reclasses a raster map greater or less than "
                "user specified area size (in hectares).",
                "module": "r.reclass.area",
            },
            "r.univar": {
                "keywords": "raster,statistics,univariate statistics,zonal "
                "statistics",
                "description": "Calculates univariate statistics from the "
                "non-null cells of a raster map. Statistics include number of "
                "cells counted, minimum and maximum cell values, "
                "range, arithmetic mean, population variance, standard "
                "deviation, coefficient of variation, and sum.",
                "module": "r.univar",
            },
            "r.covar": {
                "keywords": "raster,statistics",
                "description": "Outputs a covariance/correlation matrix for "
                "user-specified raster map layer(s).",
                "module": "r.covar",
            },
            "r.statistics": {
                "keywords": "raster,statistics,zonal statistics",
                "description": "Calculates category or object oriented "
                "statistics.",
                "module": "r.statistics",
            },
            "r.report": {
                "keywords": "raster,statistics",
                "description": "Reports statistics for raster maps.",
                "module": "r.report",
            },
            "r.mode": {
                "keywords": "raster,statistics,algebra",
                "description": "Finds the mode of values in a cover map within"
                " areas assigned the same category value in a user-specified "
                "base map.",
                "module": "r.mode",
            },
            "r.regression.line": {
                "keywords": "raster,statistics,regression",
                "description": "Calculates linear regression from two raster "
                "maps: y = a + b*x.",
                "module": "r.regression.line",
            },
            "r.regression.multi": {
                "keywords": "raster,statistics,regression",
                "description": "Calculates multiple linear regression from "
                "raster maps.",
                "module": "r.regression.multi",
            },
            "r.quant": {
                "keywords": "raster,statistics,quantization",
                "description": "Produces the quantization file for a "
                "floating-point map.",
                "module": "r.quant",
            },
            "r.stats": {
                "keywords": "raster,statistics",
                "description": "Generates area statistics for raster map.",
                "module": "r.stats",
            },
            "r.stats.quantile": {
                "keywords": "raster,statistics",
                "description": "Compute category quantiles using two passes.",
                "module": "r.stats.quantile",
            },
            "r.kappa": {
                "keywords": "raster,statistics,classification",
                "description": "Calculates error matrix and kappa parameter "
                "for accuracy assessment of classification result.",
                "module": "r.kappa",
            },
        },
        "rescale": {
            "r.rescale.eq": {
                "keywords": "raster,rescale",
                "description": "Rescales histogram equalized the range of "
                "category values in a raster map layer.",
                "module": "r.rescale.eq",
            },
            "r.rescale": {
                "keywords": "raster,rescale",
                "description": "Rescales the range of category values in a "
                "raster map layer.",
                "module": "r.rescale",
            },
        },
        "map management": {
            "r.compress": {
                "keywords": "raster,map management,compression",
                "description": "Compresses and decompresses raster maps.",
                "module": "r.compress",
            }
        },
        "color table": {
            "r.colors": {
                "keywords": "raster,color table",
                "description": "Creates/modifies the color table associated "
                "with a raster map.",
                "module": "r.colors",
            }
        },
        "composite": {
            "r.composite": {
                "keywords": "raster,composite,RGB",
                "description": "Combines red, green and blue raster maps into"
                " a single composite raster map.",
                "module": "r.composite",
            },
            "r.blend": {
                "keywords": "raster,composite",
                "description": "Blends color components of two raster maps by"
                " a given ratio.",
                "module": "r.blend",
            },
        },
        "latitude": {
            "r.latlong": {
                "keywords": "raster,latitude,longitude,projection",
                "description": "Creates a latitude/longitude raster map.",
                "module": "r.latlong",
            }
        },
        "metadata": {
            "r.describe": {
                "keywords": "raster,metadata",
                "description": "Prints terse list of category values found in"
                " a raster map layer.",
                "module": "r.describe",
            },
            "r.info": {
                "keywords": "raster,metadata,extent,history",
                "description": "Outputs basic information about a raster map.",
                "module": "r.info",
            },
            "r.timestamp": {
                "keywords": "raster,metadata,timestamp,time",
                "description": "Modifies a timestamp for a raster map. "
                "Print/add/remove a timestamp for a raster map.",
                "module": "r.timestamp",
            },
            "r.support": {
                "keywords": "raster,metadata",
                "description": "Allows creation and/or modification of raster"
                " map layer support files.",
                "module": "r.support",
            },
            "r.region": {
                "keywords": "raster,metadata",
                "description": "Sets the boundary definitions for a raster "
                "map.",
                "module": "r.region",
            },
        },
        "profile": {
            "r.profile": {
                "keywords": "raster,profile",
                "description": "Outputs the raster map layer values lying on "
                "user-defined line(s).",
                "module": "r.profile",
            }
        },
        "elevation": {
            "r.fillnulls": {
                "keywords": "raster,elevation,interpolation",
                "description": "Fills no-data areas in raster maps using "
                "spline interpolation.",
                "module": "r.fillnulls",
            },
            "r.plane": {
                "keywords": "raster,elevation",
                "description": "Creates raster plane map given dip "
                "(inclination), aspect (azimuth) and one point.",
                "module": "r.plane",
            },
            "r.relief": {
                "keywords": "raster,elevation,relief,terrain,hillshade",
                "description": "Creates shaded relief map from an elevation "
                "map (DEM).",
                "module": "r.relief",
            },
            "r.shade": {
                "keywords": "raster,elevation,relief,hillshade,visualization",
                "description": "Drapes a color raster over an shaded relief or"
                " aspect map.",
                "module": "r.shade",
            },
        },
        "algebra": {
            "r.mapcalc": {
                "keywords": "raster,algebra",
                "description": "Raster map calculator.",
                "module": "r.mapcalc",
            },
            "r.quantile": {
                "keywords": "raster,algebra,statistics",
                "description": "Compute quantiles using two passes.",
                "module": "r.quantile",
            },
            "r.mfilter": {
                "keywords": "raster,algebra,statistics,filter",
                "description": "Performs raster map matrix filter.",
                "module": "r.mfilter",
            },
            "r.neighbors": {
                "keywords": "raster,algebra,statistics,aggregation,neighbor,"
                "focal statistics,filter",
                "description": "Makes each cell category value a function of "
                "the category values assigned to the cells around it, "
                "and stores new cell values in an output raster "
                "map layer.",
                "module": "r.neighbors",
            },
            "r.texture": {
                "keywords": "raster,algebra,statistics,texture",
                "description": "Generate images with textural features from a "
                "raster map.",
                "module": "r.texture",
            },
        },
        "fire": {
            "r.spreadpath": {
                "keywords": "raster,fire,cumulative costs",
                "description": "Recursively traces the least cost path "
                "backwards to cells from which the cumulative cost was "
                "determined.",
                "module": "r.spreadpath",
            },
            "r.spread": {
                "keywords": "raster,fire,spread,hazard,model",
                "description": "Simulates elliptically anisotropic spread. "
                "Generates a raster map of the cumulative time of spread, "
                "given raster maps containing the rates of spread (ROS), the "
                "ROS directions and the spread origins. It optionally "
                "produces raster maps to contain backlink UTM "
                "coordinates for tracing spread paths. Usable for "
                "fire spread simulations.",
                "module": "r.spread",
            },
            "r.ros": {
                "keywords": "raster,fire,spread,rate of spread,hazard,model",
                "description": "Generates rate of spread raster maps. "
                "Generates three, or four raster map layers showing the base "
                "(perpendicular) rate of spread (ROS), the maximum "
                "(forward) ROS, the direction of the maximum ROS, "
                "and optionally the maximum potential spotting "
                "distance for fire spread simulation.",
                "module": "r.ros",
            },
        },
        "resample": {
            "r.resamp.filter": {
                "keywords": "raster,resample,kernel filter",
                "description": "Resamples raster map layers using an analytic "
                "kernel.",
                "module": "r.resamp.filter",
            },
            "r.resamp.rst": {
                "keywords": "raster,resample",
                "description": "Reinterpolates and optionally computes "
                "topographic analysis from input raster map to a new raster "
                "map (possibly with different resolution) using "
                "regularized spline with tension and smoothing.",
                "module": "r.resamp.rst",
            },
            "r.resample": {
                "keywords": "raster,resample",
                "description": "GRASS raster map layer data resampling "
                "capability.",
                "module": "r.resample",
            },
            "r.resamp.stats": {
                "keywords": "raster,resample",
                "description": "Resamples raster map layers to a coarser grid "
                "using aggregation.",
                "module": "r.resamp.stats",
            },
            "r.resamp.interp": {
                "keywords": "raster,resample,interpolation",
                "description": "Resamples raster map to a finer grid using "
                "interpolation.",
                "module": "r.resamp.interp",
            },
        },
        "aggregation": {
            "r.series": {
                "keywords": "raster,aggregation,series",
                "description": "Makes each output cell value a function of the"
                " values assigned to the corresponding cells in "
                "the input raster map layers.",
                "module": "r.series",
            }
        },
        "volume": {
            "r.volume": {
                "keywords": "raster,volume,clumps",
                "description": 'Calculates the volume of data "clumps". '
                "Optionally produces a GRASS vector points map containing the "
                "calculated centroids of these clumps.",
                "module": "r.volume",
            }
        },
        "buffer": {
            "r.circle": {
                "keywords": "raster,buffer,geometry,circle",
                "description": "Creates a raster map containing concentric "
                "rings around a given point.",
                "module": "r.circle",
            },
            "r.buffer": {
                "keywords": "raster,buffer",
                "description": "Creates a raster map showing buffer zones "
                "surrounding cells that contain non-NULL category values.",
                "module": "r.buffer",
            },
            "r.buffer.lowmem": {
                "keywords": "raster,buffer",
                "description": "Creates a raster map showing buffer zones "
                "surrounding cells that contain non-NULL category values. "
                "This is the low-memory alternative to the classic r.buffer "
                "module.",
                "module": "r.buffer.lowmem",
            },
        },
        "solar": {
            "r.horizon": {
                "keywords": "raster,solar,sun position",
                "description": "Computes horizon angle height from a digital "
                "elevation model. The module has two different modes of "
                "operation: 1. Computes the entire horizon around a "
                "single point whose coordinates are given with the "
                "'coord' option. The horizon height (in radians). "
                "2. Computes one or more raster maps of the horizon "
                "height in a single direction. The input for this "
                "is the angle "
                "(in degrees), which is measured "
                "counterclockwise with east=0, north=90 etc. "
                "The output is the horizon height in radians.",
                "module": "r.horizon",
            },
            "r.sunmask": {
                "keywords": "raster,solar,sun position,shadow",
                "description": "Calculates cast shadow areas from sun position"
                " and elevation raster map. Either exact sun position (A) "
                "is specified, or date/time to calculate the sun "
                "position (B) by r.sunmask itself.",
                "module": "r.sunmask",
            },
            "r.sunhours": {
                "keywords": "raster,solar,sun energy,sun position",
                "description": "Calculates solar elevation, solar azimuth, and"
                " sun hours. Solar elevation: the angle between the "
                "direction of the geometric center of the sun's "
                "apparent disk and the (idealized) horizon. Solar "
                "azimuth: the angle from due north in clockwise "
                "direction.",
                "module": "r.sunhours",
            },
            "r.sun": {
                "keywords": "raster,solar,sun energy,shadow",
                "description": "Solar irradiance and irradiation model. "
                "Computes direct (beam), diffuse and reflected solar "
                "irradiation raster maps for given day, latitude, "
                "surface and atmospheric conditions. Solar parameters "
                "(e.g. sunrise, sunset times, declination, "
                "extraterrestrial irradiance, daylight length) are "
                "saved in the map history file. Alternatively, a local "
                "time can be specified to compute solar incidence angle "
                "and/or irradiance raster maps. The shadowing effect "
                "of the topography is optionally incorporated.",
                "module": "r.sun",
            },
        },
        "viewshed": {
            "r.viewshed": {
                "keywords": "raster,viewshed,line of sight,LOS",
                "description": "Computes the viewshed of a point on an "
                "elevation raster map. Default format: NULL (invisible), "
                "vertical angle wrt viewpoint (visible).",
                "module": "r.viewshed",
            }
        },
        "querying": {
            "r.what": {
                "keywords": "raster,querying,position",
                "description": "Queries raster maps on their category values "
                "and category labels.",
                "module": "r.what",
            },
            "r.what.color": {
                "keywords": "raster,querying,color table",
                "description": "Queries colors for a raster map layer.",
                "module": "r.what.color",
            },
        },
        "null data": {
            "r.null": {
                "keywords": "raster,null data",
                "description": "Manages NULL-values of given raster map.",
                "module": "r.null",
            }
        },
        "distance": {
            "r.grow.distance": {
                "keywords": "raster,distance,proximity",
                "description": "Generates a raster map containing distances "
                "to nearest raster features.",
                "module": "r.grow.distance",
            },
            "r.distance": {
                "keywords": "raster,distance",
                "description": "Locates the closest points between objects in "
                "two raster maps.",
                "module": "r.distance",
            },
            "r.grow": {
                "keywords": "raster,distance,proximity",
                "description": "Generates a raster map layer with contiguous "
                "areas grown by one cell.",
                "module": "r.grow",
            },
        },
        "hydrology": {
            "r.carve": {
                "keywords": "raster,hydrology",
                "description": "Generates stream channels. Takes vector stream"
                " data, transforms it to raster and subtracts depth from the "
                "output DEM.",
                "module": "r.carve",
            },
            "r.uslek": {
                "keywords": "raster,hydrology,soil,erosion",
                "description": "Computes USLE Soil Erodibility Factor (K).",
                "module": "r.uslek",
            },
            "r.topmodel": {
                "keywords": "raster,hydrology,model",
                "description": "Simulates TOPMODEL which is a physically based"
                " hydrologic model.",
                "module": "r.topmodel",
            },
            "r.flow": {
                "keywords": "raster,hydrology",
                "description": "Constructs flowlines. Computes flowlines, "
                "flowpath lengths, and flowaccumulation (contributing areas) "
                "from a elevation raster map.",
                "module": "r.flow",
            },
            "r.lake": {
                "keywords": "raster,hydrology,hazard,flood",
                "description": "Fills lake at given point to given level.",
                "module": "r.lake",
            },
            "r.drain": {
                "keywords": "raster,hydrology,cost surface",
                "description": "Traces a flow through an elevation model or "
                "cost surface on a raster map.",
                "module": "r.drain",
            },
            "r.sim.sediment": {
                "keywords": "raster,hydrology,soil,sediment flow,erosion,"
                "deposition,model",
                "description": "Sediment transport and erosion/deposition "
                "simulation using path sampling method (SIMWE).",
                "module": "r.sim.sediment",
            },
            "r.terraflow": {
                "keywords": "raster,hydrology,flow,accumulation,sink",
                "description": "Performs flow computation for massive grids. "
                "Float version.",
                "module": "r.terraflow",
            },
            "r.stream.extract": {
                "keywords": "raster,hydrology,stream network",
                "description": "Performs stream network extraction.",
                "module": "r.stream.extract",
            },
            "r.watershed": {
                "keywords": "raster,hydrology,watershed",
                "description": "Calculates hydrological parameters and RUSLE "
                "factors.",
                "module": "r.watershed",
            },
            "r.fill.dir": {
                "keywords": "raster,hydrology",
                "description": "Filters and generates a depressionless "
                "elevation map and a flow direction map from a given elevation"
                " raster map.",
                "module": "r.fill.dir",
            },
            "r.water.outlet": {
                "keywords": "raster,hydrology,watershed",
                "description": "Creates watershed basins from a drainage "
                "direction map.",
                "module": "r.water.outlet",
            },
            "r.solute.transport": {
                "keywords": "raster,hydrology,solute transport",
                "description": "Numerical calculation program for transient, "
                "confined and unconfined solute transport in two dimensions",
                "module": "r.solute.transport",
            },
            "r.usler": {
                "keywords": "raster,hydrology,rainfall,soil,erosion",
                "description": "Computes USLE R factor, Rainfall erosivity "
                "index.",
                "module": "r.usler",
            },
            "r.sim.water": {
                "keywords": "raster,hydrology,soil,flow,overland flow,model",
                "description": "Overland flow hydrologic simulation using path"
                " sampling method (SIMWE).",
                "module": "r.sim.water",
            },
            "r.topidx": {
                "keywords": "raster,hydrology,wetness",
                "description": "Creates a topographic index (wetness index) "
                "raster map from an elevation raster map.",
                "module": "r.topidx",
            },
        },
        "geometry": {
            "r.patch": {
                "keywords": "raster,geometry,mosaicking,merge,patching",
                "description": "Creates a composite raster map layer by using "
                "known category values from one (or more) map layer(s) to "
                'fill in areas of "no data" in another map layer.',
                "module": "r.patch",
            },
            "r.thin": {
                "keywords": "raster,geometry",
                "description": "Thins non-null cells that denote linear "
                "features in a raster map layer.",
                "module": "r.thin",
            },
        },
        "mask": {
            "r.mask": {
                "keywords": "raster,mask",
                "description": "Creates a MASK for limiting raster operation.",
                "module": "r.mask",
            }
        },
        "color transformation": {
            "r.his": {
                "keywords": "raster,color transformation,RGB,HIS,IHS",
                "description": "Generates red, green and blue (RGB) raster map"
                " layers combining hue, intensity and saturation (HIS) values "
                "from user-specified input raster map layers.",
                "module": "r.his",
            }
        },
        "geomorphology": {
            "r.param.scale": {
                "keywords": "raster,geomorphology,terrain,elevation,landform",
                "description": "Extracts terrain parameters from a DEM. Uses a"
                " multi-scale approach by taking fitting quadratic parameters "
                "to any size window (via least squares).",
                "module": "r.param.scale",
            }
        },
    },
    "general": {
        "support": {
            "g.version": {
                "keywords": "general,support,citing,copyright,version,license",
                "description": "Displays GRASS GIS version info. Optionally "
                "also prints build or copyright information.",
                "module": "g.version",
            }
        },
        "settings": {
            "g.mapset": {
                "keywords": "general,settings",
                "description": "Changes/reports current mapset. Optionally "
                "create new mapset or list available mapsets in given "
                "location.",
                "module": "g.mapset",
            },
            "g.region": {
                "keywords": "general,settings",
                "description": "Manages the boundary definitions for the "
                "geographic region.",
                "module": "g.region",
            },
            "g.gisenv": {
                "keywords": "general,settings,variables,scripts",
                "description": "Outputs and modifies the user's current GRASS "
                "variable settings. Prints all defined GRASS variables if no "
                "option is given.",
                "module": "g.gisenv",
            },
        },
        "projection": {
            "g.proj": {
                "keywords": "general,projection,create location",
                "description": "Prints or modifies GRASS projection "
                "information files (in various coordinate system descriptions"
                "). Can also be used to create new GRASS locations.",
                "module": "g.proj",
            }
        },
        "map management": {
            "g.list": {
                "keywords": "general,map management,list",
                "description": "Lists available GRASS data base files of the "
                "user-specified data type optionally using the search "
                "pattern.",
                "module": "g.list",
            },
            "g.findfile": {
                "keywords": "general,map management,scripts",
                "description": "Searches for GRASS data base files and sets "
                "variables for the shell.",
                "module": "g.findfile",
            },
            "g.remove": {
                "keywords": "general,map management,remove",
                "description": "Removes data base element files from the "
                "user's current mapset using the search pattern.",
                "module": "g.remove",
            },
            "g.rename": {
                "keywords": "general,map management,rename",
                "description": "Renames data base element files in the user's "
                "current mapset.",
                "module": "g.rename",
            },
        },
    },
}



# -*- coding: utf-8 -*-
#######
# actinia-core - an open source REST API for scalable, distributed, high
# performance processing of geographical data that uses GRASS GIS for
# computational tasks. For details, see https://actinia.mundialis.de/
#
# Copyright (c) 2023 mundialis GmbH & Co. KG
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#######

"""
Utils needed for mapsets merge to rename the mapset inside the mapset
directory for groups, virtual rasters and tgis
"""
import fileinput
import os
import sqlite3

from actinia_core.core.common.exceptions import AsyncProcessError

__license__ = "GPLv3"
__author__ = "Anika Weinmann, Lina Krisztian"
__copyright__ = "Copyright 2023, mundialis GmbH & Co. KG"
__maintainer__ = "mundialis"


def _change_mapsetname_in_raster_vrt(
    source_path, source_mapset, target_mapset
):
    """Replaces the mapset name in the vrt files of a mapset

    Args:
        source_path(str): path of the cell_misc folder in the source mapset
        source_mapset(str): name of source mapset
        target_mapset(str): name of target mapset
    """
    for raster in os.listdir(source_path):
        vrt_file = os.path.join(source_path, raster, "vrt")
        if os.path.isfile(vrt_file):
            for line in fileinput.input(vrt_file, inplace=True):
                print(
                    line.replace(f"@{source_mapset}", f"@{target_mapset}"),
                    end="",
                )


def _change_mapsetname_in_group(group_path, source_mapset, target_mapset):
    """Replaces the mapset name in the group file

    Args:
        group_path(str): path of the group folder in the source mapset
        source_mapset(str): name of source mapset
        target_mapset(str): name of target mapset

    Raises:
        This method will raise an AsyncProcessError if a group has no REF
        file
    """
    group_dirs = os.listdir(group_path)
    for group_dir in group_dirs:
        group_file = os.path.join(group_path, group_dir, "REF")
        if os.path.isfile(group_file):
            for line in fileinput.input(group_file, inplace=True):
                print(line.replace(source_mapset, target_mapset), end="")
        else:
            raise AsyncProcessError("group %s has no REF file" % (group_dir))


def _merge_tgis_dbs(tgis_db_path_1, tgis_db_path_2):
    """Merge two tgis sqlite.db files

    Args:
        tgis_db_path_1(str): path of a tgis sqlite.db file in which the
                                other should be merged
        tgis_db_path_2(str): path of a tgis sqlite.db file which should be
                                merged in tgis_db_path_1
    """
    con = sqlite3.connect(tgis_db_path_1)
    con.execute(f"ATTACH '{tgis_db_path_2}' as dba")
    con.execute("BEGIN")

    table_names1 = [
        row[1]
        for row in con.execute(
            "SELECT * FROM sqlite_master where type='table'"
        )
    ]
    table_names2 = [
        row[1]
        for row in con.execute(
            "SELECT * FROM dba.sqlite_master where type='table'"
        )
    ]

    # merge databases
    for table in table_names2:
        if table == "tgis_metadata":
            con.execute(f"DROP TABLE {table}")
            con.execute(
                f"CREATE TABLE {table} AS " f"SELECT * FROM dba.{table}"
            )
            continue
        # for example raster_register_xxx tables are not in both dbs
        if table not in table_names1:
            con.execute(
                f"CREATE TABLE {table} AS " f"SELECT * FROM dba.{table}"
            )
            continue
        combine = f"INSERT OR IGNORE INTO {table} SELECT * FROM dba.{table}"
        con.execute(combine)
    con.commit()
    con.execute("detach database dba")
    if con:
        con.close()


def _update_views_in_tgis(tgis_db_path):
    """Update views in tgis sqlite.db

    Args:
        tgis_db_path(str): Path to the tgis sqlite.db file where the views
                            should be updated
    """
    con = sqlite3.connect(tgis_db_path)
    cur = con.cursor()

    sql_script_folder = os.path.join(os.getenv("GISBASE"), "etc", "sql")
    drop_view_sql = os.path.join(sql_script_folder, "drop_views.sql")
    with open(drop_view_sql, "r") as sql:
        sql_drop_str = sql.read()
    cur.executescript(sql_drop_str)

    view_sql_file_names = [
        "raster_views.sql",
        "raster3d_views.sql",
        "vector_views.sql",
        "strds_views.sql",
        "str3ds_views.sql",
        "stvds_views.sql",
    ]
    for view_sql_file_name in view_sql_file_names:
        view_sql_file = os.path.join(sql_script_folder, view_sql_file_name)
        with open(view_sql_file, "r") as sql:
            sql_view_str = sql.read()
        cur.executescript(sql_view_str)
    con.commit()
    if con:
        con.close()
    del cur


def _change_mapsetname_in_tgis(
    tgis_path, source_mapset, target_mapset, target_tgis_db=None
):
    """Replaces the mapset name in the tgis sqlite.db

    Args:
        tgis_path(str): path of the tgis folder in the source mapset
        source_mapset(str): name of source mapset
        target_mapset(str): name of target mapset
        target_tgis_db(str): path to existing tgis sqlite.db of target
                                mapset. None if nonexistent.
    """

    tgis_db_path = os.path.join(tgis_path, "sqlite.db")

    # tables
    con = sqlite3.connect(tgis_db_path)
    cur = con.cursor()
    table_names = [
        row[1]
        for row in cur.execute(
            "SELECT * FROM sqlite_master where type='table'"
        )
    ]
    for table_name in table_names:
        columns = [
            row[0]
            for row in cur.execute(f"SELECT * FROM {table_name}").description
        ]
        for col in columns:
            cur.execute(
                f"UPDATE {table_name} SET {col} = REPLACE({col}, "
                f"'{source_mapset}', '{target_mapset}')"
            )
    con.commit()
    if con:
        con.close()
    del cur

    # if there already exists a sqlite.db file then merge it
    if target_tgis_db is not None:
        _merge_tgis_dbs(tgis_db_path, target_tgis_db)

    # update views
    _update_views_in_tgis(tgis_db_path)


def change_mapsetname(
    source_path,
    directory,
    source_mapset,
    target_mapset,
    target_path=None,
):
    if os.path.exists(source_path) is True:
        if directory == "group":
            _change_mapsetname_in_group(
                source_path, source_mapset, target_mapset
            )
        elif directory == "tgis":
            target_tgis_db = None
            if target_path and os.path.isdir(
                os.path.join(target_path, "tgis")
            ):
                target_tgis_db = os.path.join(target_path, "tgis", "sqlite.db")
            else:
                target_path = None
            _change_mapsetname_in_tgis(
                source_path,
                source_mapset,
                target_mapset,
                target_tgis_db,
            )
        # for raster VRTs the mapset has to be changed
        elif directory == "cell_misc":
            _change_mapsetname_in_raster_vrt(
                source_path, source_mapset, target_mapset
            )



# -*- coding: utf-8 -*-
#######
# actinia-core - an open source REST API for scalable, distributed, high
# performance processing of geographical data that uses GRASS GIS for
# computational tasks. For details, see https://actinia.mundialis.de/
#
# Copyright (c) 2016-2022 Sören Gebbert and mundialis GmbH & Co. KG
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#######

"""
Asynchronous merging of several mapsets into a single one
"""
import pickle
from flask import jsonify, make_response

from actinia_core.rest.base.resource_base import ResourceBase
from actinia_core.core.common.redis_interface import enqueue_job
from actinia_core.processing.common.persistent_mapset_merger import start_job

__license__ = "GPLv3"
__author__ = "Sören Gebbert"
__copyright__ = (
    "Copyright 2016-2022, Sören Gebbert and mundialis GmbH & Co. KG"
)
__maintainer__ = "mundialis"


class AsyncPersistentMapsetMergerResource(ResourceBase):
    def __init__(self):
        ResourceBase.__init__(self)

    def post(self, location_name, mapset_name):
        """Merge several existing mapsets into a single one.
        All mapsets that should be merged and the target mapset will be locked
        for the processing.

        Args:
            location_name (str): The name of the location
            target_mapset_name (str): The name of the target mapset, into other
                                      mapsets should be merged

        Process arguments must be provided as JSON document in the POST
        request:
            ["mapset_A", "mapset_B", "mapset_C", "mapset_D", ...]

        Returns:
            flask.Response:
            The HTTP status and a JSON document that includes the
            status URL of the task that must be polled for updates.

        Example::

            {
              "HTTP code": 200,
              "Messages": "Resource accepted",
              "Resource id": "resource_id-985164c9-1db9-49cf-"
              "b2c4-3e8e48500e31",
              "Status": "accepted",
              "URLs": {
                "Resources": [],
                "Status": "http://104.155.60.87/resources/soeren/"
                          "resource_id-985164c9-1db9-49cf-b2c4-3e8e48500e31"
              },
              "User id": "soeren"
            }


        """
        # Preprocess the post call
        rdc = self.preprocess(
            has_json=True, location_name=location_name, mapset_name=mapset_name
        )

        if rdc:
            enqueue_job(self.job_timeout, start_job, rdc)

        html_code, response_model = pickle.loads(self.response_data)
        return make_response(jsonify(response_model), html_code)



# -*- coding: utf-8 -*-
#######
# actinia-core - an open source REST API for scalable, distributed, high
# performance processing of geographical data that uses GRASS GIS for
# computational tasks. For details, see https://actinia.mundialis.de/
#
# Copyright (c) 2016-2018 Sören Gebbert and mundialis GmbH & Co. KG
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#######

"""
User management

TODO: User update must be implemented
"""

from passlib.apps import custom_app_context as pwd_context
import jwt
from datetime import datetime, timezone, timedelta
from actinia_core.core.common.config import global_config
from actinia_core.core.redis_user import redis_user_interface
from actinia_core.core.common.user_base import (
    ActiniaUserBase,
)

__author__ = "Sören Gebbert, Anika Weinmann"
__copyright__ = "Copyright 2016-2022, mundialis GmbH & Co. KG"
__maintainer__ = "mundialis GmbH & Co. KG"


class ActiniaUser(ActiniaUserBase):
    """
    The Actinia Core user management class

    This class manages the user which are stored in the
    Redis database
    """

    db = redis_user_interface

    def read_from_db(self):
        creds = self.db.get_credentials(self.user_id)
        self.user_role = self.get_role()
        self.user_group = self.get_group()
        self.password_hash = self.get_password_hash()

        self.permissions = creds["permissions"]
        self.cell_limit = creds["permissions"]["cell_limit"]
        self.accessible_datasets = creds["permissions"]["accessible_datasets"]
        self.accessible_modules = creds["permissions"]["accessible_modules"]
        self.process_num_limit = creds["permissions"]["process_num_limit"]
        self.process_time_limit = creds["permissions"]["process_time_limit"]

    def __str__(self):
        creds = self.get_credentials()

        string = ""

        for entry in creds:
            if entry == "password_hash":
                continue
            string += entry + ": " + str(creds[entry]) + "\n"

        return string

    def exists(self):
        """Check if the user exists

        Returns:
            bool:
            True if the user exists, False otherwise
        """
        if self.user_id is None:
            return False

        return self.db.exists(self.user_id)

    def verify_password(self, password):
        """
        Verify the provided password with the stored passord hash

        Args:
            password (str):

        Returns:
            bool:
            True if success, False otherwise
        """
        return pwd_context.verify(password, self.get_password_hash())

    def get_role(self):
        """Return the role from the database

        Returns:
            str:
            Return the role from the database
        """
        return self.db.get_role(self.user_id)

    def get_group(self):
        """Return the user group from the database

        Returns:
            str:
            Return the user group from the database
        """
        return self.db.get_group(self.user_id)

    def get_credentials(self):
        """Return the user credentials as a dictionary

        Returns:
            dict:
            Return the user credentials as a dictionary
        """
        return self.db.get_credentials(self.user_id)

    def get_accessible_datasets(self):
        """Return a dictionary of location:mapset list entries

        Returns:
            dict:
            Return a dictionary of location:mapset list entries
        """

        self.permissions = self.db.get_credentials(self.user_id)["permissions"]

        if self.permissions and "accessible_datasets" in self.permissions:
            return self.permissions["accessible_datasets"]

    def get_accessible_modules(self):
        """Return a list of all accessible modules

        Returns:
            list:
            Return a list of all accessible modules
        """

        self.permissions = self.db.get_credentials(self.user_id)["permissions"]

        if self.permissions and "accessible_modules" in self.permissions:
            return self.permissions["accessible_modules"]

    def get_cell_limit(self):
        """Return the cell limit

        Returns:
            int:
            The value or None if nothing was found
        """

        self.permissions = self.db.get_credentials(self.user_id)["permissions"]

        if self.permissions and "cell_limit" in self.permissions:
            return self.permissions["cell_limit"]

    def get_process_num_limit(self):
        """Return the process number limit

        Returns:
            int:
            The value or None if nothing was found
        """

        self.permissions = self.db.get_credentials(self.user_id)["permissions"]

        if self.permissions and "process_num_limit" in self.permissions:
            return self.permissions["process_num_limit"]

    def get_process_time_limit(self):
        """Return the process time limit

        Returns:
            int:
            The value or None if nothing was found
        """

        self.permissions = self.db.get_credentials(self.user_id)["permissions"]

        if self.permissions and "process_time_limit" in self.permissions:
            return self.permissions["process_time_limit"]

    def get_password_hash(self):
        """Return the password hash from the database

        Returns:
            int:
            Return the password hash from the database
        """
        return self.db.get_password_hash(self.user_id)

    def generate_api_key(self):
        """Generate an API key based on the user id

        Returns:
            str:
            API key
        """
        api_key = jwt.encode(
            {"user_id": self.user_id},
            global_config.SECRET_KEY,
            algorithm="HS512",
        )

        return api_key

    def generate_auth_token(self, expiration=86400):
        """Generate an authentication token with a specific expiration time

        Args:
            expiration ()int: Time in seconds, default 86400s -> one day

        Returns:
            str:
            The auth token
        """
        exp_date = datetime.now(tz=timezone.utc) + timedelta(
            seconds=expiration
        )
        token = jwt.encode(
            {"user_id": self.user_id, "exp": exp_date},
            global_config.SECRET_KEY,
            algorithm="HS512",
        )
        return token

    def commit(self):
        """Commit the user to the database

        All internal variables must be initialized:

            self.user_id
            self.user_group
            self.password_hash
            self.user_role

        Returns:
            bool:
            True in case of success, False otherwise
        """
        if self.user_id is None:
            return False
        if self.user_group is None:
            return False
        if self.password_hash is None:
            return False
        if self.user_role is None:
            return False

        self._generate_permission_dict()

        ret = self.db.add(
            user_id=self.user_id,
            user_group=self.user_group,
            password_hash=self.password_hash,
            user_role=self.user_role,
            permissions=self.permissions,
        )
        return ret

    def update(self):
        """Update the user to the database

        This variable mus be initialized beforehand:

            self.user_id

        Only the internal variables that should be updates must be initialized:

            self.user_group
            self.password_hash
            self.user_role
            self.permissions

        Returns:
            bool:
            True in case of success, False otherwise
        """
        if self.user_id is None:
            return False

        self._generate_permission_dict()

        ret = self.db.update(
            user_id=self.user_id,
            user_group=self.user_group,
            password_hash=self.password_hash,
            user_role=self.user_role,
            permissions=self.permissions,
        )
        return ret

    def hash_password(self, password):
        """Create a password hash and store it in the object

        Args:
            password (str):

        """
        self.password_hash = pwd_context.hash(password)

    def list_all_users(self):
        """List all users in the database

        Returns:
            list:
            List of all users
        """
        return self.db.list_all_ids()

    def delete(self):
        """
        Delete a user from the user database

        Returns:
            bool:
            True if success, False otherwise
        """

        if self.exists():
            return self.db.delete(self.user_id)

        return False

    @staticmethod
    def verify_api_key(api_key):
        """Verify an API key based on the user name

        Returns:
            Actinia Core_api.common.user.ActiniaUser:
            A user object is success or None
        """
        try:
            data = jwt.decode(
                api_key,
                global_config.SECRET_KEY,
                leeway=timedelta(seconds=10),
                algorithms=["HS512"],
            )
        except jwt.exceptions.DecodeError:
            return None

        if data is None:
            return None
        if "user_id" not in data.keys():
            return None

        user = ActiniaUser(data["user_id"])
        if user.exists():
            return user

        return None

    @staticmethod
    def verify_auth_token(token):
        try:
            data = jwt.decode(
                token,
                global_config.SECRET_KEY,
                leeway=timedelta(seconds=10),
                algorithms=["HS512"],
                options={"require": ["exp"], "verify_exp": True},
            )
        except jwt.exceptions.DecodeError:
            return None
        except jwt.exceptions.ExpiredSignatureError:
            return None

        if data is None:
            return None
        if "user_id" not in data.keys():
            return None

        user = ActiniaUser(data["user_id"])
        if user.exists():
            return user
        return None

    @staticmethod
    def create_user(
        user_id,
        user_group,
        password,
        user_role="user",
        accessible_datasets={
            "nc_spm_08": ["PERMANENT", "user1", "landsat"],
            "ECAD": ["PERMANENT"],
            "latlong_wgs84": ["PERMANENT"],
        },
        accessible_modules=global_config.MODULE_ALLOW_LIST,
        cell_limit=global_config.MAX_CELL_LIMIT,
        process_num_limit=global_config.PROCESS_NUM_LIMIT,
        process_time_limit=global_config.PROCESS_TIME_LIMT,
    ):
        """Create a new user object and initialize it

        Args:
            user_id (str): The id (name, email, ..) of the user that must be
                           unique
            user_group (str): The group of the user
            user_role (str): The password
            accessible_datasets (dict): The user role (admin, user, guest)
            accessible_modules (list): Dict of location:mapset lists
            cell_limit (int): Maximum number of cells to process
            process_num_limit (int): The maximum number of processes the user
                                     is allowed to run in a single chain
            process_time_limit (int): The maximum number of seconds a user
                                      process is allowed to run

        Returns:
            actinia_core_api.common.user.ActiniaUser:
            A new user object in case of success, or None in case of failure

        """
        user = ActiniaUser(
            user_id=user_id,
            user_group=user_group,
            user_role=user_role,
            accessible_datasets=accessible_datasets,
            accessible_modules=accessible_modules,
            cell_limit=cell_limit,
            process_num_limit=process_num_limit,
            process_time_limit=process_time_limit,
        )
        user.hash_password(password)

        if user.commit() is True:
            return user
        return None



# -*- coding: utf-8 -*-
#######
# actinia-core - an open source REST API for scalable, distributed, high
# performance processing of geographical data that uses GRASS GIS for
# computational tasks. For details, see https://actinia.mundialis.de/
#
# Copyright (c) 2016-2018 Sören Gebbert and mundialis GmbH & Co. KG
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#######

"""
This module is responsible to answer requests for file based resources.
"""
from flask import jsonify, make_response
from flask_restful import Resource
from flask import send_from_directory
import os
from actinia_core.core.common.config import global_config
from actinia_core.core.common.app import auth
from actinia_core.core.common.api_logger import log_api_call
from actinia_core.core.utils import ensure_valid_path


__license__ = "GPLv3"
__author__ = "Sören Gebbert"
__copyright__ = (
    "Copyright 2016-2018, Sören Gebbert and mundialis GmbH & Co. KG"
)
__maintainer__ = "Sören Gebbert"
__email__ = "soerengebbert@googlemail.com"


class RequestStreamerResource(Resource):
    """
    This class is responsible to answer requests
    for file based resources and to stream them.
    """

    decorators = [log_api_call, auth.login_required]

    def get(self, user_id, resource_id, file_name):
        """Get the file based resource as HTTP attachment

        Args:
            user_id (str): The unique user name/id
            resource_id (str): The id of the resource
            file_name (str): The name of the file to send as attachment

        Returns:
            flask.Response: A HTTP response with the
                            file as attachment in case of success or
                            JSON payload containing the error status
                            and a message

            The HTTP status 200 header::

                Content-Disposition: attachment; filename=my_accumulation.tiff
                Content-Length: 3469
                Content-Type: image/tiff
                Last-Modified: Tue, 07 Jun 2016 10:34:17 GMT
                Cache-Control: public, max-age=43200
                Expires: Tue, 07 Jun 2016 22:34:18 GMT
                ETag: "1465295657.53-3469-665590421"
                Date: Tue, 07 Jun 2016 10:34:18 GMT


            The HTTP status 400 response JSON contents::

                {
                  "Messages": "Resource does not exist",
                  "Status": "error"
                }

        """

        resource_dir = global_config.GRASS_RESOURCE_DIR
        user_export_path = ensure_valid_path([resource_dir, user_id], "w")
        resource_export_path = ensure_valid_path(
            [user_export_path, resource_id], "w"
        )
        resource_export_file_path = ensure_valid_path(
            [resource_export_path, file_name], "w"
        )

        if (
            os.path.exists(resource_export_file_path) is True
            and os.access(resource_export_file_path, os.R_OK) is True
        ):
            return send_from_directory(
                resource_export_path, file_name, as_attachment=True
            )
        else:
            return make_response(
                jsonify(
                    {"status": "error", "message": "Resource does not exist"}
                ),
                400,
            )



# -*- coding: utf-8 -*-
#######
# actinia-core - an open source REST API for scalable, distributed, high
# performance processing of geographical data that uses GRASS GIS for
# computational tasks. For details, see https://actinia.mundialis.de/
#
# Copyright (c) 2016-2022 Sören Gebbert and mundialis GmbH & Co. KG
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#######

"""
Utils for processing
"""

import importlib
from actinia_core.core.common.config import global_config

__license__ = "GPLv3"
__author__ = "Carmen Tawalika"
__copyright__ = (
    "Copyright 2016-2022, Sören Gebbert and mundialis GmbH & Co. KG"
)
__maintainer__ = "mundialis"


def try_import(actinia_module, actinia_class):
    """Tries to import module, if module not found, throw exception
    if local queue is used.
    This is used to import processing classes. When job receiver
    cannot import it and redis queue is used, this is not a problem.
    When job receiver cannot import it and local queue is used, processing
    wouldn't work and the exception is thrown.
    Job executor needs to have it installed either way.

    Args:
        actinia_module (string): processing module to use

    Raises:
        e: ImportError or subclass

    Returns:
        module: python module used for processing
    """
    try:
        imported_module = importlib.import_module(actinia_module, package=None)
        imported_class = getattr(imported_module, actinia_class)
        return imported_class
    except ImportError as e:
        # ModuleNotFoundError is a subclass and caught here as well.
        if global_config.QUEUE_TYPE == "local":
            e.msg += (
                " - No actinia_processing found but required "
                + "for local queue!"
            )
            raise e



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#######
# actinia-core - an open source REST API for scalable, distributed, high
# performance processing of geographical data that uses GRASS GIS for
# computational tasks. For details, see https://actinia.mundialis.de/
#
# Copyright (c) 2016-2023 Sören Gebbert and mundialis GmbH & Co. KG
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#######

"""
User authentication and credential check
"""
import os
from functools import wraps

from flask import g, abort
from actinia_core.core.common.config import global_config
from actinia_core.core.common.app import auth
from actinia_core.core.common.keycloak_user import ActiniaKeycloakUser
from actinia_core.core.common.user import ActiniaUser
from actinia_core.core.common.user_noauth import ActiniaUserNoAuth
from actinia_core.core.messages_logger import MessageLogger

__license__ = "GPLv3"
__author__ = "Sören Gebbert, Julia Haas, Anika Weinmann"
__copyright__ = (
    "Copyright 2016-2023, Sören Gebbert and mundialis GmbH & Co. KG"
)
__maintainer__ = "mundialis"


if not global_config.AUTHENTICATION:
    # No authentication
    @auth.login_required
    def login_required(a, b):
        return False

    @auth.verify_password
    def verify_password(username_or_token, password):
        g.user = ActiniaUserNoAuth.create_user()
        return True


if global_config.KEYCLOAK_CONFIG_PATH:

    @auth.verify_token
    def verify_token(token):
        """Verify the keycloak token.
        A keycloak authentication token has to be provided.
        This function is called by the
        @auth.login_required decorator.
        Args:
            token (str): An authentication token
        Returns:
            bool: True if authorized or False if not
        """
        user = ActiniaKeycloakUser.verify_keycloak_token(token)
        if not user:
            return False
        g.user = user
        return True


if global_config.KEYCLOAK_CONFIG_PATH is None and global_config.AUTHENTICATION:

    @auth.verify_password
    def verify_password(username_or_token, password):
        """Verify the user name and password.

        Instead of a user name an authentication token
        or an API token can be provided.
        This function is called by the
        @auth.login_required decorator.

        Args:
            username_or_token (str): The username or an authentication token
            password (str): The optional user password, not required in case of
                            token

        Returns:
            bool: True if authorized or False if not

        """
        # first try to authenticate by token
        user = ActiniaUser.verify_auth_token(username_or_token)

        if not user:
            user = ActiniaUser.verify_api_key(username_or_token)

        if not user:
            # try to authenticate with username/password
            user = ActiniaUser(user_id=username_or_token)
            if not user.exists() or not user.verify_password(password):
                return False
        # Store the user globally
        g.user = user
        return True


def create_dummy_user(f):
    """
    Create a dummy user for configuration in which authorization is not
    required

    Args:
        f (function): The function to wrap

    Returns:
        function: The decorator functions

    """

    @wraps(f)
    def decorated_function(*args, **kwargs):
        # Set a default user
        g.user = ActiniaUser(
            user_id=global_config.DEFAULT_USER,
            user_group=global_config.DEFAULT_USER_GROUP,
        )

        return f(*args, **kwargs)

    return decorated_function


def check_admin_role(f):
    """Verify if the user has admin rights

    This decorator function verifies if the role
    of the login user is "admin" or "superadmin".

    It will abort with a 401 response if the user is not admin or superadmin

    Args:
        f (function): The function to wrap

    Returns:
        function: The decorator functions

    """

    @wraps(f)
    def decorated_function(*args, **kwargs):
        if g.user is None:
            abort(401)

        if g.user.get_role() != "admin" and g.user.get_role() != "superadmin":
            abort(401)

        return f(*args, **kwargs)

    return decorated_function


def check_user_role(f):
    """Verify if the user has user or admin rights

    This decorator function verifies if the role
    of the login user is "user", "admin" or "superadmin".

    It will abort with a 401 response if the user is not user,
    admin or superadmin

    Args:
        f (function): The function to wrap

    Returns:
        function: The decorator functions

    """

    @wraps(f)
    def decorated_function(*args, **kwargs):
        if g.user is None:
            abort(401)

        allowed_roles = ["user", "admin", "superadmin"]
        if g.user.get_role() not in allowed_roles:
            abort(401)

        return f(*args, **kwargs)

    return decorated_function


def check_admin_role_or_own_userid(f):
    """Verify if the user has admin rights or requests own user ID

    This decorator function verifies if the role
    of the login user is "admin" or "superadmin"
    or the user request the own user ID.

    It will abort with a 401 response if the user is not admin and the user
    is not the own user ID

    Args:
        f (function): The function to wrap

    Returns:
        function: The decorator functions

    """

    @wraps(f)
    def decorated_function(*args, **kwargs):
        if g.user is None:
            abort(401)

        if g.user.get_role() != "admin" and g.user.get_role() != "superadmin":
            if g.user.user_id != kwargs["user_id"]:
                abort(401)

        return f(*args, **kwargs)

    return decorated_function


def very_superadmin_role(f):
    """Verify if the user has superadmin rights

    This decorator function verifies if the role of
    of the login user is "superadmin".

    It will abort with a 401 response if the user is not superadmin

    Args:
        f (function): The function to wrap

    Returns:
        function: The decorator functions

    """

    @wraps(f)
    def decorated_function(*args, **kwargs):
        if g.user is None:
            abort(401)

        if g.user.get_role() != "superadmin":
            abort(401)

        return f(*args, **kwargs)

    return decorated_function


def check_user_permissions(f):
    """Check the user permissions

    This decorator function verifies the user permissions
    to access locations, mapsets and modules.

    The function arguments are checked if they contain:

        - location_name
        - mapset_name
        - module_name

    The user role admin can access everything.

    Args:
        f (function): The function to wrap

    Returns:
        function: The decorator functions

    """

    @wraps(f)
    def decorated_function(*args, **kwargs):
        location_name = None
        mapset_name = None
        module_name = None

        if "location_name" in kwargs:
            location_name = kwargs["location_name"]

        if "mapset_name" in kwargs:
            mapset_name = kwargs["mapset_name"]

        if "module_name" in kwargs:
            module_name = kwargs["module_name"]

        ret = check_location_mapset_module_access(
            user_credentials=g.user.get_credentials(),
            config=global_config,
            location_name=location_name,
            mapset_name=mapset_name,
            module_name=module_name,
        )
        if ret is not None:
            message_logger = MessageLogger()
            message_logger.error(str(ret[1]))
            abort(ret[0], str(ret[1]))

        return f(*args, **kwargs)

    return decorated_function


def check_location_mapset_module_access(
    user_credentials,
    config,
    location_name=None,
    mapset_name=None,
    module_name=None,
):
    """Check the user permissions to access locations, mapsets and modules.

    If the user has an admin or superadmin role, the tests are skipped.

    Args:
        user_credentials (dict): The user credentials dictionary
        config (actinia_core.core.common.config.Configuration): The actinia
                                                                configuration
        location_name (str): Name of the location to access
        mapset_name (str): Name of the mapset to access
        module_name (str): Name of the module to access

    Returns:
        tuple:
        In case of missing permissions a tuple with HTTP status
        code and a dict with status and message entries
        or "None" if the user has all required permissions or is admin

    """

    # Admin is allowed to do anything
    if (
        user_credentials["user_role"] == "admin"
        or user_credentials["user_role"] == "superadmin"
    ):
        return None

    # Mapset without location results in error
    if location_name is None and mapset_name is not None:
        resp = {
            "Status": "error",
            "Messages": "Internal error, mapset definition without location",
        }
        return (500, resp)

    if location_name:
        # Check if the location exists in the global database, if not return
        grass_data_base = config.GRASS_DATABASE
        location_path = os.path.join(grass_data_base, location_name)
        if (
            os.path.exists(location_path) is False
            or os.path.isdir(location_path) is False
            or os.access(location_path, os.R_OK & os.X_OK) is False
        ):
            return None

        # Check if the mapset exists in the global location, if not return
        if mapset_name:
            mapset_path = os.path.join(location_path, mapset_name)

            if (
                os.path.exists(mapset_path) is False
                or os.path.isdir(mapset_path) is False
                or os.access(mapset_path, os.R_OK & os.X_OK) is False
            ):
                return None

        # Check permissions to the global database locations and mapsets
        accessible_datasets = user_credentials["permissions"][
            "accessible_datasets"
        ]

        if location_name not in accessible_datasets:
            resp = {
                "Status": "error",
                "Messages": "Unauthorized access to location <%s>"
                % location_name,
            }
            return (401, resp)

        # Check if the mapset is allowed to be accessed
        if mapset_name:
            # Check if the mapset exists in the global database
            if (
                not accessible_datasets[location_name]
                or mapset_name not in accessible_datasets[location_name]
            ):
                resp = {
                    "Status": "error",
                    "Messages": "Unauthorized access to mapset "
                    "<%s> in location <%s>" % (mapset_name, location_name),
                }
                return (401, resp)

    # Check if the module name is in the access list
    if module_name:
        accessible_modules = user_credentials["permissions"][
            "accessible_modules"
        ]
        if module_name not in accessible_modules:
            resp = {
                "Status": "error",
                "Messages": "Module <%s> is not supported" % module_name,
            }
            return (401, resp)

    return None



# -*- coding: utf-8 -*-
#######
# actinia-core - an open source REST API for scalable, distributed, high
# performance processing of geographical data that uses GRASS GIS for
# computational tasks. For details, see https://actinia.mundialis.de/
#
# Copyright (c) 2016-2018 Sören Gebbert and mundialis GmbH & Co. KG
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#######

"""
Redis server interface for API logging
"""

__license__ = "GPLv3"
__author__ = "Sören Gebbert"
__copyright__ = (
    "Copyright 2016-2018, Sören Gebbert and mundialis GmbH & Co. KG"
)
__maintainer__ = "Sören Gebbert"
__email__ = "soerengebbert@googlemail.com"


class Process(object):
    """This class specifies a process object that will be executed as part of a
    process chain
    """

    def __init__(
        self,
        exec_type,
        executable,
        executable_params,
        stdin_source=None,
        skip_permission_check=False,
        id=None,
    ):
        """

        Args:
            exec_type (str): The executable type, can be "grass" or "exec"
            executable (str): The name and path of the executable, eg:
                              g.version
            or /bin/cp executable_params (list): A list of parameters (strings)
            for the executable
            stdin_source (str): The get_stdout or get_stderr method of a
                                Process
            skip_permission_check(boolean): Skip permission check for the
                                            module or executable, this is
                                            meaningful for internal process
                                            chain use. Hence the user can use
                                            internal process chains that
                                            contain module he has no
                                            permissions to use.
            id (str): The unique id of the process
        """

        self.exec_type = exec_type
        self.executable = executable
        if isinstance(self.executable, bytes):
            self.executable = self.executable.decode()

        self.executable_params = executable_params
        self.stdin_source = stdin_source
        self.stdout = None
        self.stderr = None
        self.skip_permission_check = skip_permission_check
        self.id = id

    def set_stdouts(self, stdout, stderr):
        """Set the content of stdout and stderr of this process

        Set this after the process has finished.

        Args:
            stdout: The stdout string of this process
            stderr: The stderr string of this process
        """
        self.stdout = stdout
        self.stderr = stderr

    def get_stdout(self):
        return self.stdout

    def get_stderr(self):
        return self.stderr

    def __str__(self):
        return (
            self.exec_type
            + " "
            + self.executable
            + " "
            + str(self.executable_params)
        )
